use crate::config::RollBotConfig;
use crate::roll_bot::RollBot;
use log::{error, info, warn};
use prost::Message;
use raas_types::raas::bot::roll::{roll_cmd, Roll, RollCmd, RollImage, RollResponse};
use raas_types::raas::cmd::request::Cmd;
use raas_types::raas::cmd::Command;
use raas_types::raas::ping::{Ping, Pong};
use raas_types::raas::register::{BotTypes, Register, RegisterResponse};
use raas_types::raas::resp::response::Resp;
use raas_types::raas::resp::Response;
use raas_types::raas::{recv_raas_msg, send_raas_msg};
use std::net::TcpStream;
use std::time::{SystemTime, UNIX_EPOCH};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Roll Bot Error")]
    RollBot(#[from] crate::roll_bot::Error),
    #[error("Network Error")]
    IO(#[from] std::io::Error),
    #[error("Protobuf Encode Error")]
    ProtoBufEncode(#[from] prost::EncodeError),
    #[error("Protobuf Decode Error")]
    ProtoBufDecode(#[from] prost::DecodeError),
    #[error("Invalid command ID")]
    InvalidCommandID,
    #[error("Invalid command")]
    InvalidCommand,
}

pub struct Client {
    config: RollBotConfig,
    roll_bot: RollBot,

    id: Option<u32>,
}

impl Client {
    pub fn new(config: RollBotConfig, roll_bot: RollBot) -> Self {
        Self {
            config,
            roll_bot,
            id: None,
        }
    }

    fn try_connect(&self) -> Result<TcpStream, Error> {
        for attempt in 1..10 {
            info!("Attempting to connect to {}", self.config.server_addr);
            let socket = TcpStream::connect(self.config.server_addr.clone());

            if let Ok(socket) = socket {
                return Ok(socket);
            }

            let wait = attempt * 10;
            warn!("Connection failed, waiting {}s...", wait);
            std::thread::sleep(std::time::Duration::from_secs(wait))
        }

        Err(std::io::Error::last_os_error().into())
    }

    fn register(&mut self, socket: &mut TcpStream) -> Result<(), Error> {
        let register_msg = Register {
            name: "Roll Bot".to_string(),
            bot_type: BotTypes::Roll.into(),
        };

        let mut message = Vec::new();
        register_msg.encode(&mut message)?;

        send_raas_msg(socket, message)?;

        let resp = recv_raas_msg(socket)?;

        let register_msg = RegisterResponse::decode(&*resp.msg)?;

        info!(
            "Registered '{}' as id {}",
            register_msg.name, register_msg.id
        );

        self.id = Some(register_msg.id);

        Ok(())
    }

    fn get_next_command(&mut self, socket: &mut TcpStream) -> Result<RollCmd, Error> {
        let msg = recv_raas_msg(socket)?;

        let command = Command::decode(&*msg.msg)?;

        info!(
            "Received command for id={} timestamp={}",
            command.id,
            &command.request.as_ref().unwrap().timestamp
        );

        if command.id != self.id.unwrap() {
            warn!("Command for a different ID was received, dropping");
            return Err(Error::InvalidCommandID);
        }

        match command.request.unwrap().cmd.unwrap() {
            Cmd::RollCmd(cmd) => Ok(cmd),
            #[allow(unreachable_patterns)]
            _ => {
                warn!("Invalid command id received");
                Err(Error::InvalidCommand)
            }
        }
    }

    fn handle_ping(&mut self, ping: Ping) -> Result<RollResponse, Error> {
        info!("Received ping w/ data {:?}", ping.data);

        let pong = Pong {
            ping: Some(ping),
            is_ok: true,
        };

        let resp = RollResponse {
            response: Some(raas_types::raas::bot::roll::roll_response::Response::Pong(
                pong,
            )),
        };

        Ok(resp)
    }

    fn handle_roll(&mut self, roll: Roll) -> Result<RollResponse, Error> {
        info!("Doing roll...");
        self.roll_bot.roll(roll.rotations)?;

        let capture_exit = std::process::Command::new("libcamera-still")
            .arg("--output")
            .arg("/tmp/out.jpg")
            .status()
            .unwrap();

        let img_data = if !capture_exit.success() {
            error!("Failed to capture image ret={}", capture_exit);
            vec![]
        } else {
            std::fs::read("/tmp/out.jpg")?
        };

        let resp = RollResponse {
            response: Some(
                raas_types::raas::bot::roll::roll_response::Response::RollImage(RollImage {
                    img: img_data,
                }),
            ),
        };

        Ok(resp)
    }

    fn handle_command(&mut self, socket: &mut TcpStream, roll_cmd: RollCmd) -> Result<(), Error> {
        let roll_resp = match roll_cmd.cmd.unwrap() {
            roll_cmd::Cmd::Ping(ping) => self.handle_ping(ping),
            roll_cmd::Cmd::Roll(roll) => self.handle_roll(roll),
        }?;

        let timestamp = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .unwrap()
            .as_secs();

        let resp = Response {
            id: self.id.unwrap(),
            timestamp,
            resp: Some(Resp::RollResp(roll_resp)),
        };

        let mut message = Vec::new();

        resp.encode(&mut message)?;

        send_raas_msg(socket, message)?;

        Ok(())
    }

    pub fn run_client(&mut self) -> Result<(), Error> {
        loop {
            let mut socket = self.try_connect()?;

            self.register(&mut socket)?;

            loop {
                let roll_cmd_res = self.get_next_command(&mut socket);

                if let Ok(roll_cmd) = roll_cmd_res {
                    self.handle_command(&mut socket, roll_cmd)?;
                } else if let Err(roll_cmd_err) = roll_cmd_res {
                    match roll_cmd_err {
                        Error::IO(io_error) => {
                            warn!("Connection issue, reattempting connection: {}", io_error);
                            self.id = None;
                            break;
                        }
                        Error::InvalidCommandID | Error::InvalidCommand => {
                            continue;
                        }
                        _ => return Err(roll_cmd_err),
                    }
                }
            }
        }
    }
}
